﻿local L = AceLibrary("AceLocale-2.2"):new("sRaidFrames")
local BS = AceLibrary("Babble-Spell-2.2")
local surface = AceLibrary("Surface-1.0")
local roster = AceLibrary("Roster-2.1")
local Aura = AceLibrary("SpecialEvents-Aura-2.0")

surface:Register("Otravi", "Interface\\AddOns\\sRaidFrames\\textures\\otravi")
surface:Register("Smooth", "Interface\\AddOns\\sRaidFrames\\textures\\smooth")
surface:Register("Striped", "Interface\\AddOns\\sRaidFrames\\textures\\striped")
surface:Register("BantoBar", "Interface\\AddOns\\sRaidFrames\\textures\\bantobar")

local math_mod = math.fmod or math.mod

sRaidFrames = AceLibrary("AceAddon-2.0"):new(
	"AceDB-2.0",
	"AceEvent-2.0",
	"AceConsole-2.0",
	"AceHook-2.1",
	"FuBarPlugin-2.0"
)

-- Look I'm a fubar plugin!!
sRaidFrames.hasIcon = "Interface\\Icons\\INV_Helmet_06"
sRaidFrames.defaultMinimapPosition = 180
sRaidFrames.cannotDetachTooltip = true
sRaidFrames.hasNoColor = true
sRaidFrames.clickableTooltip = false
sRaidFrames.hideWithoutStandby = true
sRaidFrames.independentProfile = true

-- bindings
BINDING_NAME_ShowHideRaidWindows = L["Show/Hide Raid Windows"]
BINDING_NAME_ToggleSort = L["Toggle Sort"]
BINDING_NAME_Sortbyclass = L["Sort by class"]
BINDING_NAME_Sortbygroup = L["Sort by group"]
BINDING_NAME_ToggleBuffDebuffview = L["Toggle Buff/Debuff view"]
BINDING_NAME_Toggledisplayofonlydispellabledebuffs = L["Toggle display of only dispellable debuffs"]

local function NormalizeSpellName(name)
	return string.gsub(name, "[^%a]", "")
end

function sRaidFrames:OnInitialize()

	self:RegisterDB("sRaidFramesDB")
	self:Variables()

	self:RegisterDefaults("profile", {
		Locked				= false,
		SortBy				= "class",
		HealthFormat	= 'percent',
		Invert 				= false,
		Scale					= 1,
		Border				= true,
		Texture				= "Otravi",
		BuffType			= "debuffs",
		ShowOnlyDispellable	= 1,
		BackgroundColor		= {r = 0.32, g = 0.50, b = 0.70, a = 0.90},
		BorderColor		= {r = 1, g = 1, b = 1, a = 1},
		Growth				= "down",
		Spacing				= 0,
		ShowGroupTitles		= true,
		SubSort				= "name",
		UnitTooltipMethod		= "notincombat",
		BuffTooltipMethod = "always",
		DebuffTooltipMethod = "always",
		ClassFilter			= {["WARRIOR"] = true, ["PALADIN"] = true, ["SHAMAN"] = true, ["HUNTER"] = true, ["WARLOCK"] = true, ["MAGE"] = true, ["DRUID"] = true, ["ROGUE"] = true, ["PRIEST"] = true},
		GroupFilter			= {true, true, true, true, true, true, true, true},
		BuffFilter			= {},
		PowerFilter			= {[0] = true,[1] = true,[2] = true,[3] = true},
		RangeCheck 			= false,
		RangeLimit			= 40,
		RangeFrequency 		= 2,
		RangeAlpha 			= 0.5,
		AggroCheck				= false,
		HighlightTarget	= false,
		HighlightHeals	= true,
		DebuffColors		= { ["Curse"] = { ["r"] = 1, ["g"] = 0, ["b"] = 0.75, ["a"] = 0.5},
												["Magic"]    = { ["r"] = 1, ["g"] = 0, ["b"] = 0, ["a"] = 0.5},
												["Disease"]  = { ["r"] = 1, ["g"] = 1, ["b"] = 0, ["a"] = 0.5},
												["Poison"]   = { ["r"] = 0, ["g"] = 0.5, ["b"] = 0, ["a"] = 0.5} },
		Layout					= "CTRA_WithBorders",
	})

	self:RegisterChatCommand({"/srf", "/sraidframes"}, self.options)

	self.opt = self.db.profile

	self.OnMenuRequest = self.options

	self.master = CreateFrame("Frame", "sRaidFrame", UIParent)
	self.master:SetMovable(true)
	self.master:SetScale(self.opt.Scale)

	self.master:SetHeight(200);
	self.master:SetWidth(200);
end

function sRaidFrames:OnProfileEnable()
	self.opt = self.db.profile
	self:SetPosition()
end

function sRaidFrames:OnEnable()
	self:chatUpdateBuffMenu()

	self:SecureHook("SecureRaidGroupHeader_Update", "SecureHeaderUpdate")
	self:RegisterBucketEvent("RAID_ROSTER_UPDATE", 0.2, "UpdateRoster")
	self:RegisterEvent("LEARNED_SPELL_IN_TAB", "ScanSpellbookForRange")
	-- self:RegisterEvent("LEARNED_SPELL_IN_TAB", "ScanSpellbookForDispells") -- TODO, doesn't work
	self:RegisterEvent("RosterLib_UnitChanged")

	for i = 1, 8 do
		self:CreateGroupFrame(i)
	end

	self:SetGroupFilters()
	self:SetSortMethods()
	self:SetGrowth()
	self:SetPosition()

	self:AddRangeFunction(10, function (unit) return CheckInteractDistance(unit, 3) == 1 end)
	self:AddRangeFunction(28, function (unit) return CheckInteractDistance(unit, 4) == 1 end)
	self:AddRangeFunction(100, function (unit) return UnitIsVisible(unit) == 1 end)

	self:ScanSpellbookForRange()
	self:ScanSpellbookForDispells()

	self:UpdateRoster()
end

function sRaidFrames:OnDisable()
	self.enabled = false
	self.master:Hide()
end

function sRaidFrames:JoinedRaid()
	self.enabled = true
	self.statusstate = {}

	self:RegisterBucketEvent("UNIT_HEALTH", 0.2)
	self:RegisterBucketEvent("UNIT_AURA", 0.2)

	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UpdateTarget")
	
	if CastCommLib then CastCommLib:RegisterCallback(self, "CastCommCallback") end

	self:RegisterEvent("Banzai_UnitGainedAggro")
	self:RegisterEvent("Banzai_UnitLostAggro")

	self:RegisterEvent("oRA_PlayerCanResurrect")
	self:RegisterEvent("oRA_PlayerResurrected")
	self:RegisterEvent("oRA_PlayerNotResurrected")
	
	self:RegisterEvent("SpecialEvents_UnitDebuffGained")
	self:RegisterEvent("SpecialEvents_UnitDebuffLost")
	self:RegisterEvent("SpecialEvents_UnitBuffGained")
	self:RegisterEvent("SpecialEvents_UnitBuffLost")

	self:RegisterBucketEvent({"UNIT_MANA", "UNIT_RAGE", "UNIT_ENERGY"}, 1, "UNIT_HEALTH")
	self:ScheduleRepeatingEvent("sRaidFramesRangeCheck", self.RangeCheck, self.opt.RangeFrequency, self)

	self:UpdateRoster()
	self:UpdateAll()

	self.master:Show()
end

function sRaidFrames:LeftRaid()
	self.visible = {}
	self.statusstate = {}
	self.enabled = false

	self:UnregisterBucketEvent("UNIT_HEALTH")
	self:UnregisterBucketEvent("UNIT_AURA")

	self:UnregisterEvent("Banzai_UnitGainedAggro")
	self:UnregisterEvent("Banzai_UnitLostAggro")

	self:UnregisterEvent("oRA_PlayerCanResurrect")
	self:UnregisterEvent("oRA_PlayerResurrected")
	self:UnregisterEvent("oRA_PlayerNotResurrected")

	self:CancelScheduledEvent("sRaidFramesUpdateAll")
	self:CancelScheduledEvent("sRaidFramesRangeCheck")

	if CastCommLib then CastCommLib:UnregisterCallback(self) end

	self.master:Hide()
end

function sRaidFrames:UpdateAll()
	for unit in pairs(self.visible) do
		self:UpdateUnit(unit)
		self:UpdateAuras(unit)
	end
end

function sRaidFrames:Variables()
	self.enabled = false
	self.frames, self.visible, self.groupframes = {}, {}, {}
	self.unavail, self.res, self.RangeChecks = {}, {}, {}
	self.dispellTypes = {}
	self.CastCommUser = {}

	self.cooldownSpells = {}
	self.cooldownSpells["WARLOCK"] = BS["Soulstone Resurrection"]
	self.cooldownSpells["DRUID"] = BS["Rebirth"]
	self.cooldownSpells["SHAMAN"] = BS["Reincarnation"]

	self.classGroups = {}
	self.classGroups[1] = "WARRIOR"
	self.classGroups[2] = "MAGE"
	self.classGroups[3] = (UnitFactionGroup("player") == "Alliance") and "PALADIN" or "SHAMAN"
	self.classGroups[4] = "DRUID"
	self.classGroups[5] = "HUNTER"
	self.classGroups[6] = "ROGUE"
	self.classGroups[7] = "WARLOCK"
	self.classGroups[8] = "PRIEST"

	self.eclass2class = {}
	self.eclass2class["PALADIN"] = L["Paladin"]
	self.eclass2class["WARRIOR"] = L["Warrior"]
	self.eclass2class["SHAMAN"] = L["Shaman"]
	self.eclass2class["MAGE"] = L["Mage"]
	self.eclass2class["HUNTER"] = L["Hunter"]
	self.eclass2class["DRUID"] = L["Druid"]
	self.eclass2class["ROGUE"] = L["Rogue"]
	self.eclass2class["PRIEST"] = L["Priest"]
	self.eclass2class["WARLOCK"] = L["Warlock"]


	self.cleaningSpells = {}
	self.cleaningSpells["Poison"] = {BS["Purify"], BS["Cleanse"], BS["Cure Poison"], BS["Abolish Poison"]}
	self.cleaningSpells["Disease"] = {BS["Cure Disease"], BS["Abolish Disease"], BS["Purify"], BS["Cleanse"]}
	self.cleaningSpells["Magic"] = {BS["Dispel Magic"], BS["Cleanse"], BS["Devour Magic"]}
	self.cleaningSpells["Curse"] = {BS["Remove Curse"], BS["Remove Lesser Curse"]}
	
	self.ManaBarColor = ManaBarColor

	-- Nurfed mana color, where is the love :(
	--self.ManaBarColor[0] = { r = 0.00, g = 1.00, b = 1.00, prefix = TEXT(MANA) };

	self.statusmaps = {}
	self.statusmaps["Aggro"] = {["priority"] = 50}
	self.statusmaps["Target"] = {["priority"] = 55}
	self.statusmaps["Heal"] = {["priority"] = 60}
	
	self.statusmaps["BuffText"] = {["priority"] = 50}
	self.statusmaps["Debuff_Curse"] = {["priority"] = 55}
	self.statusmaps["Debuff_Magic"] = {["priority"] = 54}
	self.statusmaps["Debuff_Disease"] = {["priority"] = 53}
	self.statusmaps["Debuff_Poison"] = {["priority"] = 52}
	
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Shield Wall"])] = {["priority"] = 53}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Last Stand"])] = {["priority"] = 52}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Vanish"])] = {["priority"] = 51}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Gift of Life"])] = {["priority"] = 51}
	
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Blessing of Protection"])] = {["priority"] = 50}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Stealth"])] = {["priority"] = 50}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Innervate"])] = {["priority"] = 50}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Ice Block"])] = {["priority"] = 50}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Divine Shield"])] = {["priority"] = 50}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Power Infusion"])] = {["priority"] = 50}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Fear Ward"])] = {["priority"] = 49}
	self.statusmaps["Buff_".. NormalizeSpellName(BS["Shadowform"])] = {["priority"] = 48}
	

	self.statusmaps["Death"] = {["priority"] = 70}
	
	self.statusElements = {"Border", "Background", "Statusbar"}
	
	self.statusstate = {}
end

function sRaidFrames:UpdateRoster()
	local num = GetNumRaidMembers()

	if num == 0 then
		if self.enabled then
			self:LeftRaid()
		end
		return
	end

	if not self.enabled then
			self:JoinedRaid()
	end
end

function sRaidFrames:UNIT_HEALTH(units)
	for unit in pairs(units) do
		self:UpdateUnit(unit)
	end
end

function sRaidFrames:UNIT_AURA(units)
	for unit in pairs(units) do
		self:UpdateAuras(unit)
	end
end

function sRaidFrames:RosterLib_UnitChanged(unitid, name, class, subgroup, rank, oldname, oldunitid, oldclass, oldsubgroup, oldrank)
	if not string.match(unitid or oldunitid, "^raid(%d+)$") then return end
	
	if not name then
		-- Unit left
		self.statusstate[oldunitid] = nil
		self.CastCommUser[oldunitid] = nil
	elseif not oldname then
		-- Unit joined
		self.visible[unitid] = true -- TODO: Legacy, remove
		self.statusstate[unitid], self.CastCommUser[unitid] = {}, nil
		self:UpdateUnit(unitid)
		self:UpdateAuras(unitid)
		self:UpdateStatuses(unitid)
	else
		-- Unit changed
		self.visible[unitid] = true -- TODO: Legacy, remove
		self.statusstate[unitid], self.CastCommUser[unitid] = {}, nil
		self:UpdateUnit(unitid)
		self:UpdateAuras(unitid)
		self:UpdateStatuses(unitid)
	end
end

local LastTarget
function sRaidFrames:UpdateTarget()
	if not self.opt.HighlightTarget then return end

	if LastTarget then
		self:UnsetStatus("Border", LastTarget, "Target")
	end

	for unit in pairs(self.visible) do
		if UnitIsUnit(unit, "target") then
			self:SetStatus("Border", unit, "Target", {r = 1, g = 0.75, b = 0}, "Target")
			LastTarget = unit
			break
		end
	end
	
end

function sRaidFrames:SpecialEvents_UnitDebuffGained(unit, debuff, apps, type, tex, rank, index)
	if not self.visible[unit] then return end
	if not type or type == "" then return end
	if self.opt.ShowOnlyDispellable and not self.dispellTypes[type] then return end -- Only show dispellable buffs, and we can't dispell it, exit
	self:SetStatus("Background", unit, "Debuff_".. type, self.opt.DebuffColors[type], debuff)
end

function sRaidFrames:SpecialEvents_UnitDebuffLost(unit, debuff, apps, type, tex, rank)
	if not self.visible[unit] then return end
	if not type or type == "" then return end
	if not Aura:UnitHasDebuffType(unit, type) then
		self:UnsetStatus("Background", unit, "Debuff_".. type)
	end
end

function sRaidFrames:SpecialEvents_UnitBuffGained(unit, buff, index, apps, tex, rank)
	if not self.visible[unit] then return end
	if buff == BS["Feign Death"] then
		self:UnsetStatus("Background", unit, "Death")
		
	elseif tex == "Interface\\Icons\\Spell_Nature_TimeStop" and buff == BS["Divine Intervention"] then
		--f.hpbar.text:SetText("|cffff0000"..L["Intervened"].."|r")
		
	elseif tex == "Interface\\Icons\\Spell_Nature_Lightning" and buff == BS["Innervate"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=0,g=1,b=0,a=1}, L["Innervating"])
		
	elseif tex == "Interface\\Icons\\Spell_Holy_GreaterHeal" and buff == BS["Spirit of Redemption"] then
		--self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=0,b=0,a=1}, L["Spirit"])
		
	elseif tex == "Interface\\Icons\\Ability_Warrior_ShieldWall" and buff == BS["Shield Wall"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, BS["Shield Wall"])
		
	elseif tex == "Interface\\Icons\\Spell_Holy_AshesToAshes" and buff == BS["Last Stand"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, BS["Last Stand"])
		
	elseif tex == "Interface\\Icons\\INV_Misc_Gem_Pearl_05" and buff == BS["Gift of Life"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, BS["Gift of Life"])
		
	elseif tex == "Interface\\Icons\\Spell_Frost_Frost" and buff == BS["Ice Block"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, BS["Ice Block"])
		
	elseif tex == "Interface\\Icons\\Spell_Holy_SealOfProtection" and buff == BS["Blessing of Protection"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, L["Protection"])
		
	elseif tex == "Interface\\Icons\\Spell_Holy_DivineIntervention" and buff == BS["Divine Shield"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, BS["Divine Shield"])
		
	elseif tex == "Interface\\Icons\\Ability_Vanish" and buffName == BS["Vanish"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, L["Vanished"])
		
	elseif tex == "Interface\\Icons\\Ability_Stealth" and buff == BS["Stealth"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, L["Stealthed"])
		
	elseif tex == "Interface\\Icons\\Spell_Holy_PowerInfusion" and buff == BS["Power Infusion"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=1,a=1}, L["Infused"])
		
	elseif tex == "Interface\\Icons\\Spell_Holy_Excorcism" and buff == BS["Fear Ward"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=1,b=0,a=1}, BS["Fear Ward"])
		
	elseif tex == "Interface\\Icons\\Spell_Shadow_Shadowform" and buff == BS["Shadowform"] then
		self:SetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff), {r=1,g=0,b=0.75,a=1}, BS["Shadowform"])
	end
end

function sRaidFrames:SpecialEvents_UnitBuffLost(unit, buff, apps, type, tex, rank)
	if not self.visible[unit] then return end
	self:UnsetStatus("Statusbar", unit, "Buff_"..NormalizeSpellName(buff))
end

function sRaidFrames:Banzai_UnitGainedAggro(unit, unitTable)
	if not unit or not self.visible[unit] or not self.opt.AggroCheck then return end

	self:SetStatus("Border", unit, "Aggro", {r = 1, g = 0, b = 0}, "Aggro")
end

function sRaidFrames:Banzai_UnitLostAggro(unit)
	if not unit or not self.visible[unit] or not self.opt.AggroCheck then return end

	self:UnsetStatus("Border", unit, "Aggro")
end

function sRaidFrames:oRA_PlayerCanResurrect(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then 
		self.res[unit] = 1 
		self:UpdateUnit(unit)
	end
end

function sRaidFrames:oRA_PlayerResurrected(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then 
		self.res[unit] = 2 
		self:UpdateUnit(unit)
	end
end

function sRaidFrames:oRA_PlayerNotResurrected(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then 
		self.res[unit] = nil
		self:UpdateUnit(unit)
	end	
end

function sRaidFrames:IsUnitInRange(unit, range)
	if not self.RangeChecks[range] then
		return false
	end
	return self.RangeChecks[range](unit)
end

function sRaidFrames:RangeCheck()
	if not self.opt.RangeCheck then return end
	local RangeLimit = self.opt.RangeLimit

	for unit in pairs(self.visible) do
		if self:IsUnitInRange(unit, RangeLimit) then
			for _, f in pairs(self:FindUnitFrames(unit)) do
				f:SetAlpha(1)
			end
		else
			for _, f in pairs(self:FindUnitFrames(unit)) do
				f:SetAlpha(self.opt.RangeAlpha)
			end
		end
	end
end

function sRaidFrames:AddRangeFunction(range, check)
	self.RangeChecks[range] = check
	self:UpdateRangeLimitOptions()
end

-- Credits to pastamancer <3
function sRaidFrames:ScanSpellbookForRange()
	local i = 1
	local Gratuity = AceLibrary("Gratuity-2.0")
	local sName, sRank, sRange

	repeat
		local sIndex = i
		sName, sRank = GetSpellName(i, BOOKTYPE_SPELL)
		if sName and IsSpellInRange(i, "spell", "player") ~= nil then
			Gratuity:SetSpell(i, BOOKTYPE_SPELL)
			_, _, sRange = Gratuity:Find(L["(%d+) yd range"], 2, 2)
			if sRange then
				self:AddRangeFunction(tonumber(sRange), function (unit) return IsSpellInRange(sIndex, "spell", unit) == 1 end)
			end
		end
		i = i + 1
	until not sName
end

function sRaidFrames:ScanSpellbookForDispells()
	local i = 1
	local sName

	repeat
		sName = GetSpellName(i, BOOKTYPE_SPELL)
		if sName then
			for type, spells in pairs(self.cleaningSpells) do
				if not self.dispellTypes[type] then
					for _, spell in pairs(spells) do
						if sName == spell then
							self.dispellTypes[type] = true
							break
						end
					end
				end
			end
		end
		i = i + 1
	until not sName
end

function sRaidFrames:UpdateRangeFrequency()
		self:CancelScheduledEvent("sRaidFramesRangeCheck")
		self:ScheduleRepeatingEvent("sRaidFramesRangeCheck", self.RangeCheck, self.opt.RangeFrequency, self)
end

function sRaidFrames:UpdateStatuses(unit)
	for _, element in pairs(self.statusElements) do
		self:UpdateStatusElement(element, unit)
	end
end

function sRaidFrames:UpdateUnit(unit)
	if self.visible[unit] and UnitExists(unit) then
		for _, f in pairs(self:FindUnitFrames(unit)) do
			local _, class = UnitClass(unit)
			if class then
				f.title:SetText(UnitName(unit))

				local color = RAID_CLASS_COLORS[class]
				f.title:SetTextColor(color.r, color.g, color.b, 1)
			else
				f.title:SetText(UnitName(unit) or L["Unknown"])
			end

			if class == "HUNTER" and Aura:UnitHasBuff(unit, BS["Feign Death"]) then
				f.hpbar.text:SetText("|cff00ff00"..BS["Feign Death"].."|r")
				f.hpbar:SetValue(100)
				f.hpbar:SetStatusBarColor(0, 0.9, 0.5)
				f.mpbar:SetValue(0)
				self:UnsetStatus("Background", unit, "Death")
			else
				local status, dead, ghost = nil, UnitIsDead(unit), UnitIsGhost(unit)

				if not UnitIsConnected(unit) then status = "|cffff0000"..L["Offline"].."|r"
				elseif self.res[unit] == 1 and dead then status = "|cffff8c00"..L["Can Recover"].."|r"
				elseif self.res[unit] == 2 and (dead or ghost) then status = "|cffff8c00"..L["Resurrected"].."|r"
				elseif ghost then status = "|cffff0000"..L["Released"].."|r"
				elseif dead then status = "|cffff0000"..L["Dead"].."|r"
				end

				if status then
					self.unavail[unit] = true
					f.hpbar.text:SetText(status)
					f.hpbar:SetValue(0)
					f.mpbar.text:SetText()
					f.mpbar:SetValue(0)
					self:SetStatus("Background", unit, "Death", {r = 0.1, g = 0.1, b = 0.1, a = 1})
				else
					self:UnsetStatus("Background", unit, "Death")
					self.unavail[unit] = false
					self.res[unit] = nil
					local hp = UnitHealth(unit) or 0
					local hpmax = UnitHealthMax(unit)
					local hpp = (hpmax ~= 0) and ceil((hp / hpmax) * 100) or 0
					local hptext, hpvalue = nil, nil

					if self.opt.HealthFormat == "percent" then
						hptext = hpp .."%"
					elseif self.opt.HealthFormat == "deficit" then
						hptext = (hp-hpmax) ~=  0 and (hp-hpmax) or nil
					elseif self.opt.HealthFormat == "current" then
						hptext = hp
					elseif self.opt.HealthFormat == "curmax" then
						hptext = hp .."/".. hpmax
					end

					if self.opt.Invert then
						hpvalue = 100 - hpp
					else
						hpvalue = hpp
					end

					f.hpbar.text:SetText(hptext)
					f.hpbar:SetValue(hpvalue)
					f.hpbar:SetStatusBarColor(self:GetHPSeverity(hpp/100))

					local powerType = UnitPowerType(unit)
					if self.opt.PowerFilter[powerType] == false then
						f.mpbar:SetValue(0)
					else
						local color = self.ManaBarColor[powerType]
						local mp = UnitMana(unit) or 0
						local mpmax = UnitManaMax(unit)
						local mpp = (mpmax ~= 0) and ceil((mp / mpmax) * 100) or 0
						f.mpbar:SetStatusBarColor(color.r, color.g, color.b)
						f.mpbar:SetValue(mpp)
					end
				end
			end
		end
	end
end

function sRaidFrames:UpdateAuras(unit)
	if self.visible[unit] and UnitExists(unit) then
		for _, f in pairs(self:FindUnitFrames(unit)) do
			for i = 1, self:GetLayout().debuffCount do
				f["aura".. i]:Hide()
			end

			for i = 1, self:GetLayout().buffCount do
				f["buff".. i]:Hide()
			end

			local debuffSlots = 0
			if (self.opt.BuffType == "debuffs" or self.opt.BuffType == "buffsifnotdebuffed") then
				for i=1,40 do
					local debuffName, _, debuffTexture, debuffApplications, debuffType = UnitDebuff(unit, i, self.opt.ShowOnlyDispellable)
					if not debuffTexture then break end
					
					debuffSlots = debuffSlots + 1
					local debuffFrame = f["aura".. debuffSlots]
					debuffFrame.unitid = unit
					debuffFrame.debuffid = i
					debuffFrame.showDispellable = self.opt.ShowOnlyDispellable
					debuffFrame.count:SetText(debuffApplications > 1 and debuffApplications or nil);
					debuffFrame.texture:SetTexture(debuffTexture)
					debuffFrame:Show()
					
					if debuffSlots == self:GetLayout().debuffCount then break end
				end
			end

			if self.opt.BuffType == "buffs" or (self.opt.BuffType == "buffsifnotdebuffed" and debuffSlots == 0) then
				local buffSlots = 0
				local showOnlyCastable = 1
				if next(self.opt.BuffFilter) then
					showOnlyCastable = 0
				end
				for i=1,32 do
					local buffName, _, buffTexture, buffApplications = UnitBuff(unit, i, showOnlyCastable)
					if not buffTexture then break end

					if showOnlyCastable == 1 or self.opt.BuffFilter[buffName] then
						buffSlots = buffSlots + 1
						local buffFrame = f["buff".. buffSlots]
						buffFrame.buffid = i
						buffFrame.unitid = unit
						buffFrame.showCastable = showOnlyCastable
						buffFrame.count:SetText(buffApplications > 1 and buffApplications or nil)
						buffFrame.texture:SetTexture(buffTexture)
						buffFrame:Show()
					end

					if buffSlots == self:GetLayout().buffCount then break end
				end
			end
		end
	end
end

function sRaidFrames:SetStatus(element, unit, name, color, text)
	if not self.visible[unit] then return end
	--self:Print("Set", element, unit, name, color, text)
	if not self.statusmaps[name] then self:Print("Unknown status", name) return end
	if not self.statusstate[unit] then
		self.statusstate[unit] = {}
	end
	
	if not self.statusstate[unit][element] then
		self.statusstate[unit][element] = {}
	end
	
	if not self.statusstate[unit][element][name] then
		self.statusstate[unit][element][name] = {}
	end
	
	
	local cache = self.statusstate[unit][element][name]
	cache.color = color
	cache.text = text
	cache.name = name
	
	self:UpdateStatusElement(element, unit)
end

function sRaidFrames:UpdateStatusElement(element, unit)
	local status = self:GetTopStatus(element, unit)
	--self:Print("TopStatus", element, unit, status)
	for _, f in pairs(self:FindUnitFrames(unit)) do
		if element == "Border" then
			if status == nil then
				f:SetBackdropBorderColor(self.opt.BorderColor.r, self.opt.BorderColor.g, self.opt.BorderColor.b, self.opt.BorderColor.a)
			else
				f:SetBackdropBorderColor(status.color.r, status.color.g, status.color.b, status.color.a)
			end
		elseif element == "Statusbar" then
			if status == nil then
				f.mpbar.text:SetText(nil)
			else
				f.mpbar.text:SetText(status.text)
				f.mpbar.text:SetTextColor(status.color.r, status.color.g, status.color.b, status.color.a)
			end		
		elseif element == "Background" then
			if status == nil then
				f:SetBackdropColor(self.opt.BackgroundColor.r, self.opt.BackgroundColor.g, self.opt.BackgroundColor.b, self.opt.BackgroundColor.a)
			else
				f:SetBackdropColor(status.color.r, status.color.g, status.color.b, status.color.a)
			end
		end
	end	
end

function sRaidFrames:GetTopStatus(element, unit)
	if not self.statusstate[unit] or not self.statusstate[unit][element] then 
		return nil
	end
	
	local TopStatus, TopPriority = nil, 0
	for name, data in pairs(self.statusstate[unit][element]) do
		if self.statusmaps[name].priority > TopPriority then
			TopPriority = self.statusmaps[name].priority
			TopStatus = data
		end
	end

	return TopStatus
end

function sRaidFrames:UnsetStatus(element, unit, name)
	if not self.visible[unit] then return end
	if self.statusmaps[name] and self.statusstate[unit] and self.statusstate[unit][element] and self.statusstate[unit][element][name] then
		self.statusstate[unit][element][name] = nil
		self:UpdateStatusElement(element, unit)
	end
end

function sRaidFrames:GetHPSeverity(percent)
	if (percent >= 0.5) then
		return (1.0-percent)*2, 1.0, 0.0
	else
		return 1.0, percent*2, 0.0
	end
end

function sRaidFrames:QueryTooltipDisplay(value)
	if value == "never" then
		return false
	elseif value == "notincombat" and UnitAffectingCombat("player") then
		return false
	else
		return true
	end
end

function sRaidFrames:UnitTooltip(frame)
	local name, rank, subgroup, level, class, fileName, zone, online, isDead = GetRaidRosterInfo(frame.id)
	GameTooltip:SetOwner(frame)
	GameTooltip:AddDoubleLine(name, level, RAID_CLASS_COLORS[fileName].r, RAID_CLASS_COLORS[fileName].g, RAID_CLASS_COLORS[fileName].b, 1, 1, 1)
	if UnitIsAFK(frame.unit) then
		GameTooltip:AddLine("AFK: Away From Keyboard", 1, 1, 0)
	end
	if UnitIsDND(frame.unit) then
		GameTooltip:AddLine("DND: Do Not Disturb", 1, 1, 0)
	end
	GameTooltip:AddLine(UnitRace(frame.unit) .. " " .. class, 1, 1, 1);
	GameTooltip:AddDoubleLine(zone, "Group ".. subgroup, 1, 1, 1, 1, 1, 1);

	if self.CastCommUser[frame.unit] then
		GameTooltip:AddLine("CastComm availiable!", 0, 1, 0)
	end

	if oRAOCoolDown and oRAOCoolDown.db.profile.cooldowns and oRAOCoolDown.db.profile.cooldowns[name] and self.cooldownSpells[fileName] then
		GameTooltip:AddLine(" ")
		local expire = oRAOCoolDown.db.profile.cooldowns[name]-time()
		if expire > 0 then
  		GameTooltip:AddDoubleLine(self.cooldownSpells[fileName], SecondsToTime(expire), nil, nil, nil, 1, 0, 0)
  	else
  		GameTooltip:AddDoubleLine(self.cooldownSpells[fileName], "Ready!", nil, nil, nil, 0, 1, 0)
  	end
	end

	GameTooltip:Show()
end

function sRaidFrames_initialHeaderConfigFunction(frame)
	frame:SetAttribute("initial-width", sRaidFrames:GetLayout().unitframeWidth)
	frame:SetAttribute("initial-height", sRaidFrames:GetLayout().unitframeHeight)
	frame:SetAttribute("type1", "target")
	frame:SetAttribute("*type1", "target")
end

function sRaidFrames_InitUnitFrame(frame)
	sRaidFrames:CreateUnitFrame(frame)
end

function sRaidFrames_OnAttributeChanged(self, name, value)
	if name == "unit" then
		if value then
			_, _, self.id = string.find(value, "(%d+)") -- TODO, don't know if I like this
			self.unit = value
			sRaidFrames:UpdateUnit(self.unit)
			sRaidFrames:UpdateAuras(self.unit)
			sRaidFrames:UpdateStatuses(self.unit)
		else
			self.id = nil
			self.unit = nil
		end
	end
end

-- Adapts frames created by Secure Headers
function sRaidFrames:CreateUnitFrame(f)
	f:EnableMouse(true)
	f:RegisterForClicks("LeftButtonUp", "RightButtonUp", "MiddleButtonUp", "Button4Up", "Button5Up")

	f:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.UnitTooltipMethod) then self:UnitTooltip(this) end end)
	f:SetScript("OnLeave", function() GameTooltip:Hide() end)

	f.title = f:CreateFontString(nil, "ARTWORK")
	f.title:SetFontObject(GameFontNormalSmall)
	f.title:SetJustifyH("LEFT")


	for i = 1, self:GetLayout().debuffCount do
		local debuffFrame = CreateFrame("Button", nil, f)
		debuffFrame:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.DebuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitDebuff(this.unitid, this.debuffid, this.showDispellable) end end);
		debuffFrame:SetScript("OnLeave", function() GameTooltip:Hide() end)
		debuffFrame.texture = debuffFrame:CreateTexture(nil, "ARTWORK")
		debuffFrame.texture:SetAllPoints(debuffFrame);
		debuffFrame.count = debuffFrame:CreateFontString(nil, "OVERLAY")
		debuffFrame.count:SetFontObject(GameFontHighlightSmallOutline)
		debuffFrame.count:SetJustifyH("CENTER")
		debuffFrame.count:SetPoint("CENTER", debuffFrame, "CENTER", 0, 0);
		debuffFrame:Hide()
		f["aura"..i] = debuffFrame
	end

	for i = 1, self:GetLayout().buffCount do
		local buffFrame = CreateFrame("Button", nil, f)
		buffFrame:SetScript("OnEnter", function() if self:QueryTooltipDisplay(self.opt.BuffTooltipMethod) then GameTooltip:SetOwner(this) GameTooltip:SetUnitBuff(this.unitid, this.buffid, this.showCastable) end end)
		buffFrame:SetScript("OnLeave", function() GameTooltip:Hide() end)
		buffFrame.texture = buffFrame:CreateTexture(nil, "ARTWORK")
		buffFrame.texture:SetAllPoints(buffFrame)
		buffFrame.count = buffFrame:CreateFontString(nil, "OVERLAY")
		buffFrame.count:SetFontObject(GameFontHighlightSmallOutline)
		buffFrame.count:SetJustifyH("CENTER")
		buffFrame.count:SetPoint("CENTER", f.buff1, "CENTER", 0, 0);
		buffFrame:Hide()
		f["buff"..i] = buffFrame
	end
	
	f.hpbar = CreateFrame("StatusBar", nil, f)
	f.hpbar:SetStatusBarTexture(surface:Fetch(self.opt.Texture))
	f.hpbar:SetMinMaxValues(0,100)

	f.hpbar.texture = f.hpbar:CreateTexture(nil, "BORDER")
	f.hpbar.texture:SetTexture(surface:Fetch(self.opt.Texture))
	f.hpbar.texture:SetVertexColor(1, 0, 0, 0)

	f.hpbar.text = f.hpbar:CreateFontString(nil, "ARTWORK")
	f.hpbar.text:SetFontObject(GameFontHighlightSmall)
	f.hpbar.text:SetJustifyH("CENTER")

	f.mpbar = CreateFrame("StatusBar", nil, f)
	f.mpbar:SetStatusBarTexture(surface:Fetch(self.opt.Texture))
	f.mpbar:SetMinMaxValues(0,100)

	f.mpbar.texture = f.mpbar:CreateTexture(nil, "BORDER")
	f.mpbar.texture:SetTexture(surface:Fetch(self.opt.Texture))
	f.mpbar.texture:SetVertexColor(1, 0, 0, 0)

	f.mpbar.text = f.mpbar:CreateFontString(nil, "ARTWORK")
	f.mpbar.text:SetFontObject(GameFontHighlightSmall)
	f.mpbar.text:SetJustifyH("CENTER")

	self:GetLayout().StyleUnitFrame(f)

	f:Hide();

	tinsert(self.frames, f)

	ClickCastFrames = ClickCastFrames or {}
	ClickCastFrames[f] = true
end

function sRaidFrames:SetGroupFilters()
	for _, f in pairs(self.groupframes) do
		local id = f:GetID()
		if self.opt.SortBy == "group" then
			f.header:SetAttribute("groupFilter", self.opt.GroupFilter[id] and id or nil)
			f.title:SetText(string.format(L["Group %d"], id))
		elseif self.opt.SortBy == "class" then
			f.header:SetAttribute("groupFilter", self.opt.ClassFilter[self.classGroups[id]] and self.classGroups[id] or nil)
			f.title:SetText(self.eclass2class[self.classGroups[id]])
		end
		f.header:Show()
	end
end

function sRaidFrames:SecureHeaderUpdate(frame)
	if frame:GetParent().anchor then
		self:ScheduleLeaveCombatAction(self.UpdateTitleVisibility, self, frame)
	end
end

function sRaidFrames:UpdateTitleVisibility(frame)
	local i = 1
	local hasChildren = false
	local child = frame:GetAttribute("child"..i)
	while child do
		if UnitExists(child.unit) and child:IsVisible() then
			hasChildren = true
 			break
		end
		i = i + 1
		child = frame:GetAttribute("child"..i)
	end

	if self.opt.ShowGroupTitles and hasChildren then
		frame:GetParent().anchor:Show()
	else
		frame:GetParent().anchor:Hide()
	end
end

function sRaidFrames:SetSortMethods()
	for _, f in pairs(self.groupframes) do
		f.header:SetAttribute("sortMethod", "NAME")
		f.header:SetAttribute("sortDir", "DESC")
	end
end

function sRaidFrames:SetGrowth()
	for _, f in pairs(self.groupframes) do
		f.header:ClearAllPoints();
		if self.opt.Growth == "up" then
			f.header:SetAttribute("point", "TOP")
			f.header:SetPoint("BOTTOM", f.anchor, "TOP")
		elseif self.opt.Growth == "right" then
			f.header:SetAttribute("point", "LEFT")
			f.header:SetPoint("LEFT", f.anchor, "RIGHT")
		elseif self.opt.Growth == "left" then
			f.header:SetAttribute("point", "RIGHT")
			f.header:SetPoint("RIGHT", f.anchor, "LEFT")
		elseif self.opt.Growth == "down" then
			f.header:SetAttribute("point", "BOTTOM")
			f.header:SetPoint("TOP", f.anchor, "BOTTOM")
		end
	end
	self:SetSpacing()
end

function sRaidFrames:SetSpacing()
	local s = sRaidFrames.opt.Spacing
	for _, f in pairs(self.groupframes) do
		if self.opt.Growth == "down" then
			f.header:SetAttribute("xOffset", 0)
			f.header:SetAttribute("yOffset", s)
		elseif self.opt.Growth == "up" then
			f.header:SetAttribute("xOffset", 0)
			f.header:SetAttribute("yOffset", -s)
		elseif self.opt.Growth == "left" then
			f.header:SetAttribute("xOffset", s)
			f.header:SetAttribute("yOffset", 0)
		elseif self.opt.Growth == "right" then
			f.header:SetAttribute("xOffset", -s)
			f.header:SetAttribute("yOffset", 0)
		end
	end
end

function sRaidFrames:StartMovingAll(f)
	this.multidrag = 1
	local id = f:GetID()
	local fg = self.groupframes[id]
	local x, y = fg:GetLeft(), fg:GetTop()
	if ( not x or not y ) then
		return
	end
	for k, f in pairs(self.groupframes) do
		if k ~= id then
			local oX, oY = f:GetLeft(), f:GetTop()
			if ( oX and oY ) then
				f:ClearAllPoints()
				f:SetPoint("TOPLEFT", fg, "TOPLEFT", oX-x, oY-y)
			end
		end
	end
end

function sRaidFrames:StopMovingOrSizingAll(f)
	this.multidrag = nil
	local id = this:GetID()
	local fg = self.groupframes[id]
	for k, f in pairs(self.groupframes) do
		if k ~= id then
			local oX, oY = f:GetLeft(), f:GetTop()
			if ( oX and oY ) then
				f:ClearAllPoints()
				f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", oX, oY)
			end
		end
	end
end

function sRaidFrames:CreateGroupFrame(id)
	local f = CreateFrame("Frame", "sRaidFramesGroupBase".. id, self.master)
	f:SetHeight(self:GetLayout().headerHeight)
	f:SetWidth(self:GetLayout().headerWidth)
	f:SetMovable(true)
	f:SetID(id)

	f.header = CreateFrame("Frame", "sRaidFramesGroupHeader" .. id, f, "SecureRaidGroupHeaderTemplate")
	f.header:SetAttribute("template", "sRaidFrameUnitSecure")
	f.header.initialConfigFunction = sRaidFrames_initialHeaderConfigFunction

	f.anchor = CreateFrame("Button", "sRaidFramesAnchor"..id, self.master)
	f.anchor:SetHeight(self:GetLayout().headerHeight)
	f.anchor:SetWidth(self:GetLayout().headerWidth)
	f.anchor:SetScript("OnDragStart", function() if self.opt.Locked then return end if IsAltKeyDown() then self:StartMovingAll(f) end f:StartMoving() end)
	f.anchor:SetScript("OnDragStop", function() if this.multidrag == 1 then self:StopMovingOrSizingAll() end f:StopMovingOrSizing(f) self:SavePosition() end)
	f.anchor:EnableMouse(true)
	f.anchor:RegisterForDrag("LeftButton")
	f.anchor:RegisterForClicks("LeftButtonUp","RightButtonUp","MiddleButtonUp","Button4Up","Button5Up")
	f.anchor:SetPoint("CENTER", f)

	f.title = f.anchor:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	f.title:SetJustifyH("CENTER")
	f.title:ClearAllPoints()
	f.title:SetPoint("CENTER", f.anchor)

	self.groupframes[id] = f
end

function sRaidFrames:SetBackdrop(f)
	if self.opt.Border then
		f:SetBackdrop({ bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
										tile = true,
										tileSize = 16,
										edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
										edgeSize = 16,
										insets = { left = 5, right = 5, top = 5, bottom = 5 }
									})
	else
		f:SetBackdrop({ bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
										tile = true, tileSize = 16,
										insets = { left = 0, right = 0, top = 0, bottom = 0 }
									})
	end

	f:SetBackdropColor(self.opt.BackgroundColor.r, self.opt.BackgroundColor.g, self.opt.BackgroundColor.b, self.opt.BackgroundColor.a)
	f:SetBackdropBorderColor(self.opt.BorderColor.r, self.opt.BorderColor.g, self.opt.BorderColor.b, self.opt.BorderColor.a)
end

function sRaidFrames:SetStyle(f)
	self:SetWHP(f, 90, 40)
	self:SetWHP(f.title, 80, 16, "TOPLEFT", f, "TOPLEFT",  5, -4)
	self:SetWHP(f.aura1, 16, 16, "TOPRIGHT", f, "TOPRIGHT", -4, -4)
	self:SetWHP(f.aura2, 16, 16, "RIGHT", f.aura1, "LEFT", 0, 0)
	self:SetWHP(f.buff1, 12, 12, "TOPRIGHT", f, "TOPRIGHT", -4, -4)
	self:SetWHP(f.buff2, 12, 12, "RIGHT", f.buff1, "LEFT", 0, 0)
	self:SetWHP(f.buff3, 12, 12, "RIGHT", f.buff2, "LEFT", 0, 0)
	self:SetWHP(f.buff4, 12, 12, "RIGHT", f.buff3, "LEFT", 0, 0)
	self:SetWHP(f.hpbar, 80, 12, "TOPLEFT", f.title, "BOTTOMLEFT", 0, 0)
	self:SetWHP(f.mpbar, 80, 4, "TOPLEFT", f.hpbar, "BOTTOMLEFT", 0, 0)

	self:SetWHP(f.hpbar.text, f.hpbar:GetWidth(), f.hpbar:GetHeight(), "CENTER", f.hpbar, "CENTER", 0, 0)
	self:SetWHP(f.mpbar.text, f.mpbar:GetWidth(), f.mpbar:GetHeight(), "CENTER", f.mpbar, "CENTER", 0, 0)

	self:SetBackdrop(f)
end

function sRaidFrames:SetWHP(frame, width, height, p1, relative, p2, x, y)
	frame:SetWidth(width)
	frame:SetHeight(height)

	if (p1) then
		frame:ClearAllPoints()
		frame:SetPoint(p1, relative, p2, x, y)
	end
end

function sRaidFrames:FindUnitFrames(unit)
	local res = {}
	for _, f in ipairs(self.frames) do
		if f.unit and f.unit == unit then
			tinsert(res, f)
		end
	end
	return res
end

function sRaidFrames:SetPosition()
	if not self.opt.Positions then
		self:ResetPosition()
	else
		self:RestorePosition()
	end
end

function sRaidFrames:SavePosition()
	local aryPos = {}
	local s = self.master:GetEffectiveScale()

	for k,f in pairs(self.groupframes) do
		aryPos[k] = {x = f:GetLeft()*s, y = f:GetTop()*s}
	end

	self:S("Positions", aryPos)
end

function sRaidFrames:RestorePosition()
	local aryPos = self.opt.Positions
	local s = self.master:GetEffectiveScale()

	for k, pos in pairs(aryPos) do
		local f = self.groupframes[k]
		f:ClearAllPoints()
		f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", pos["x"]/s, pos["y"]/s)
	end
end

function sRaidFrames:ResetPosition()
	self:PositionLayout("ctra", 200, -200)
end

function sRaidFrames:PositionLayout(layout, xBuffer, yBuffer)
	local xMod, yMod, i = 0, 0, -1
	local frameHeight = self:GetLayout().unitframeHeight+3+self.opt.Spacing

	for k,v in pairs(self.groupframes) do
		i = i + 1
		if layout == "horizontal" then
			yMod = (i) * v:GetWidth()
			xMod = 0
		elseif layout == "vertical" then
			if i ~= 0 and math_mod(i, 2) == 0 then
				xMod = xMod + (-1*MEMBERS_PER_RAID_GROUP*frameHeight)
				yMod = 0
				i = 0
			else
				yMod = i * v:GetWidth()
			end
		elseif layout == "ctra" then
			if i ~= 0 and math_mod(i, 2) == 0 then
				yMod = yMod + v:GetWidth()
				xMod = 0
				i = 0
			else
				xMod = i * (-1*MEMBERS_PER_RAID_GROUP*frameHeight)
			end
		end

		v:ClearAllPoints()
		v:SetPoint("TOPLEFT", UIParent, "TOPLEFT", xBuffer+yMod, yBuffer+xMod)
	end

	self:SavePosition()
end

local Tablet = AceLibrary("Tablet-2.0")
function sRaidFrames:OnTooltipUpdate()
	Tablet:SetHint(L["Right-click for options."])
end

function sRaidFrames:S(var, val)
	self.db.profile[var] = val
end
